import argparse
from vinca.lib import classes
from pathlib import Path
collection = classes.collection
deckdict = classes.deckdict

import datetime
TODAY = datetime.date.today()
DAY = datetime.timedelta(days=1)

# type checking
def deck_type(arg):
	if arg.isdigit(): arg = int(arg)
	assert arg in deckdict.keys(), 'Bad deck. Use [d] to list decks'
	return deckdict[arg]
def card_type(arg):
	assert arg.isdigit()
	assert int(arg) in collection.id_list, 'Bad id. Use [q] to search for a card.'
	return classes.Card(arg)
def deck_or_card(arg):
	if arg.isdigit(): arg = int(arg)
	if arg in deckdict.keys():  # can be deck's name or deck id
		return deckdict[arg]
	if arg in collection.id_list:
		return classes.Card(arg) 
	raise argparse.ArgumentTypeError(f'''\n\n
		"{arg}" is neither a deck nor a card.
		Valid arguments are:
		1) A deck name
		2) A deck id (use d to list deck ids)
		3) A card id (use q to search for a card id)''')
def date_type(arg):
	try:
		return TODAY + int(arg) * DAY
	except:
		pass  # arg cannot be interpreted as an integer
	try:
		return datetime.date(*[int(x) for x in arg.split('-')])
	except:
		raise argparse.ArgumentTypeError(f'''\n\n
			Invalid Date: {arg}. Valid dates are:
			1) -8		(relative to today)
			2) 2021-06-03	(absolute)''')

# argument parsing
parser = argparse.ArgumentParser()
parser.set_defaults(decks = [], cards = [],
	decks_or_cards = [], func = 'statistics', mode='normal')
subparsers = parser.add_subparsers()
# commands which take a deck as an argument
one_liner_parser = subparsers.add_parser('one_liner',aliases=['1'],
	help='add a basic card quickly')
one_liner_parser.add_argument('decks',type=deck_type,nargs='*')
one_liner_parser.set_defaults(func = 'one_liner')

linear_add_parser = subparsers.add_parser('linear_add',aliases=['l'],
	help='for lyrics, poetry, oratory, etc.')
linear_add_parser.add_argument('decks',type=deck_type,nargs='*')
linear_add_parser.set_defaults(func = 'linear_add')

add_parser = subparsers.add_parser('add',aliases=['a'],
	help='add a basic card')
add_parser.add_argument('decks',type=deck_type,nargs='*')
add_parser.set_defaults(func = 'add')

add_many_parser = subparsers.add_parser('add_many',aliases=['A'],
	help='add several basic cards')
add_many_parser.add_argument('decks',type=deck_type,nargs='*')
add_many_parser.set_defaults(func = 'add_many')


image_cloze_parser = subparsers.add_parser('image_cloze',aliases=['ic'],
	help='generate an image cloze card')
image_cloze_parser.add_argument('image_path',type=Path)
image_cloze_parser.add_argument('decks',type=deck_type,nargs='*')
image_cloze_parser.set_defaults(func = 'image_cloze')

query_parser = subparsers.add_parser('query',aliases=['q'],
	help='search collection for regexp')
query_parser.add_argument('pattern',nargs='?',default='')
query_parser.add_argument('-i', '--id_only',action='store_true',
	help='only output the id of the match')
query_parser.set_defaults(func = 'query')

visual_query_parser = subparsers.add_parser('visual_query',aliases=['vq'],
	help='search for regexp and perform command on the selected card')
visual_query_parser.add_argument('pattern',nargs='?',default='')
visual_query_parser.set_defaults(func = 'query', mode='visual')

filter_parser = subparsers.add_parser('filter',aliases=['f'],
	help='filter the collection')
filter_parser.add_argument('-i','--id_only',action='store_true',
	help='only output the id of the match (facilitates piping)')
# parameters which correspond to those of the browser.filter function
filter_parser.add_argument('--decks',type=deck_type,nargs='+', default=[])
filter_parser.add_argument('--cards',type=card_type,nargs='+', default=[])
filter_parser.add_argument('--tags_include',nargs='+', metavar='TAGS')
filter_parser.add_argument('--tags_exclude',nargs='+', metavar='TAGS')
filter_parser.add_argument('--create_date_min',type=date_type, metavar='DATE')
filter_parser.add_argument('--create_date_max',type=date_type, metavar='DATE')
filter_parser.add_argument('--seen_date_min',type=date_type, metavar='DATE')
filter_parser.add_argument('--seen_date_max',type=date_type, metavar='DATE')
filter_parser.add_argument('--due_date_min',type=date_type, metavar='DATE')
filter_parser.add_argument('--due_date_max',type=date_type, metavar='DATE')
filter_parser.add_argument('--due_only',action='store_true')
filter_parser.add_argument('--not_due_only',action='store_true')
filter_parser.add_argument('--editor', type=str)
filter_parser.add_argument('--reviewer', type=str)
filter_parser.add_argument('--scheduler', type=str)
filter_parser.add_argument('--deleted_only',action='store_true')
filter_parser.add_argument('--not_deleted_only',action='store_true')
filter_parser.add_argument('--new_only',action='store_true')
filter_parser.add_argument('--not_new_only',action='store_true')
filter_parser.set_defaults(func = 'filter')


visual_filter_parser = subparsers.add_parser('visual_filter',aliases=['vf'],
	help='filter the collection and select a card')
# parameters which correspond to those of the browser.filter function
visual_filter_parser.add_argument('--decks',type=deck_type,nargs='+',default=[])
visual_filter_parser.add_argument('--cards',type=card_type,nargs='+',default=[])
visual_filter_parser.add_argument('--tags_include',nargs='+', metavar='TAGS')
visual_filter_parser.add_argument('--tags_exclude',nargs='+', metavar='TAGS')
visual_filter_parser.add_argument('--create_date_min',type=date_type, metavar='DATE')
visual_filter_parser.add_argument('--create_date_max',type=date_type, metavar='DATE')
visual_filter_parser.add_argument('--seen_date_min',type=date_type, metavar='DATE')
visual_filter_parser.add_argument('--seen_date_max',type=date_type, metavar='DATE')
visual_filter_parser.add_argument('--due_date_min',type=date_type, metavar='DATE')
visual_filter_parser.add_argument('--due_date_max',type=date_type, metavar='DATE')
visual_filter_parser.add_argument('--due_only',action='store_true')
visual_filter_parser.add_argument('--not_due_only',action='store_true')
visual_filter_parser.add_argument('--editor', type=str)
visual_filter_parser.add_argument('--reviewer', type=str)
visual_filter_parser.add_argument('--scheduler', type=str)
visual_filter_parser.add_argument('--deleted_only',action='store_true')
visual_filter_parser.add_argument('--not_deleted_only',action='store_true')
visual_filter_parser.add_argument('--new_only',action='store_true')
visual_filter_parser.add_argument('--not_new_only',action='store_true')
visual_filter_parser.set_defaults(func = 'filter', mode='visual')


# commands which take a deck or several cards as an argument 
study_parser = subparsers.add_parser('study',aliases=['s'],
	help='study the collection or selected deck')
study_parser.add_argument('decks_or_cards', type=deck_or_card, nargs='*')
study_parser.add_argument('--date', type=int,
	help='study as if today was [date]')
study_parser.set_defaults(func = 'study')

statistics_parser = subparsers.add_parser('statistics',aliases=['S'],
	help='statistics about the selected deck or card')
statistics_parser.add_argument('decks_or_cards', type=deck_or_card, nargs='*')
statistics_parser.set_defaults(func = 'statistics')
# TODO miscellaneous options for more advanced statistics
edit_parser = subparsers.add_parser('edit',aliases=['e'],
	help='edit the selected deck or card')
edit_parser.add_argument('decks_or_cards',type=deck_or_card,nargs='*')
edit_parser.set_defaults(func = 'edit')

delete_parser = subparsers.add_parser('delete',aliases=['x'],
	help='delete the selected deck or card')
delete_parser.add_argument('decks_or_cards',type=deck_or_card, nargs='*')
delete_parser.set_defaults(func = 'delete')

# commands which take no arguments
list_decks_parser = subparsers.add_parser('list_decks',aliases=['d'],
	help='list decks')
list_decks_parser.set_defaults(func = 'list_decks')

visual_decks_parser = subparsers.add_parser('visual_decks',aliases=['vd'],
	help='select a deck and perform a command on it')
visual_decks_parser.set_defaults(func = 'list_decks', mode='visual')

purge_parser = subparsers.add_parser('purge',aliases=['p'],
	help='permanently delete all cards scheduled for deletion')
purge_parser.set_defaults(func = 'purge')

# commands which take a path as an argument (import / backup)
backup_parser = subparsers.add_parser('backup',aliases=['b','export'],
	help='backup all cards')
backup_parser.add_argument('backup_dest',type=Path)
backup_parser.add_argument('decks_or_cards',type=deck_or_card,nargs='*') 
backup_parser.set_defaults(func = 'backup')

import_parser = subparsers.add_parser('import',aliases=['i'],
	help='import a collection of cards')
import_parser.add_argument('import_path',type=Path)
import_parser.add_argument('-o','--overwrite',action='store_true',
	help='overwrite the existing collection')
import_parser.set_defaults(func = 'import_collection')



# The default reviewer
# The parent module (reviewers/__init__.py) contatins some
# generic code used by all of the reviewers.
from vinca.lib.ansi import ansi
import re
import readchar 

from vinca.lib.video import DisplayImage
from vinca.lib.audio import Recording
from vinca.lib.terminal import AlternateScreen

def default_str(card):
	f = ' / '.join((card.path / 'front').read_text().splitlines())
	b = ' / '.join((card.path / 'back').read_text().splitlines())
	return f + ' | ' + b

def review(card,mode):
	global spelltest_words; spelltest_words = []

	with AlternateScreen():
		# front text
		front = (card.path / 'front').read_text()
		print(re.sub(r'\{.*?\}', colorize, front))

		# front media
		with DisplayImage(card.path/'image_front'), Recording(card.path/'audio_front'):
			# card flip
			if spelltest_words:
				spelltest(spelltest_words)
			else:
				char = readchar.readchar() # press any key to flip the card
			
		# back text
		back = '\n' + (card.path / 'back').read_text()
		print(re.sub(r'\{.*?\}', colorize, back))

		# front media
		with DisplayImage(card.path/'image_back'), Recording(card.path/'audio_back'):
			key = readchar.readchar()
	
	return key

spelltest_words = []
def spelltest(spelltest_words):
	for word in spelltest_words:
		print(ansi['show_cursor'],end='')
		i = input(f'{ansi["light"]}  spell test:   {ansi["reset"]}')
		print(ansi['hide_cursor'],end='')
		print(ansi['move_up_line'] + ansi['clear_line'],end='')
	
		# evaluate input
		if i.strip().lower() == word.strip().lower():  # right
			print(f'{ansi["green"]} {word} {ansi["reset"]}')
		else:
			print(f'{ansi["red"]} {i} {ansi["reset"]}')  # wrong
			print(f'{ansi["green"]} {word} {ansi["reset"]}')

# WARNING: functional side effect
def colorize(match_obj):
	''' a basic function which uses ansi terminal codes to markup the text
	whose inputs look like {it, b, yellow, spell: word}
	It has the side effect of processing spell: attr to spelltest_words.'''
	# WARNING: functional side effect
	global spelltest_words

	s = match_obj.group()
	s = s[1:-1]  # strip off {braces}
	attrs, word = s.split(':')
	attrs = [attr.strip() for attr in attrs.split(',')]
	
	# add words to the spell test (see warning)
	if 'type' in attrs: spelltest_words.append(word); attrs.remove('type')
	if 'spell' in attrs: spelltest_words.append(word); attrs.remove('spell')

	assert all([a in ansi.keys() for a in attrs]), f'Bad Attribute\n{ansi.keys()}'
	attrs = ''.join([ansi[attr] for attr in attrs])
	return attrs + word + ansi["reset"]




import datetime
today = datetime.date.today()
import importlib

def edit(card, mode):
	start = datetime.datetime.now()

	m = importlib.import_module('.' + card.editor, package = 'vinca.editors')
	m.edit(card, mode)

	stop = datetime.datetime.now()
	elapsed_time = min(240, (stop - start).seconds)

	card.add_history(today, elapsed_time, 0)



# The metareviewer which implements specific reviewers
import datetime
today = datetime.date.today()
import importlib
	
grade_dict = {'-': -3,
	      '@': -2,
	      'x': -1,
	      'q': -1,
	      'u': -1,
	      'h': 0,
	      'w': 0,
	      '1': 1,
	      '2': 2,
	      '3': 3,
	      '4': 4,
	      ' ': 3,
	      '\r': 3,
	      '\n': 3}
cont_dict = {'-': 1,
	     '@': 1,
	     'x': 1,
	     'q': 0,
	     'u': -1,
	     'h': -1,
	     'w': 0,
	     '1': 1,
	     '2': 1,
	     '3': 1,
	     '4': 1,
	     ' ': 1,
	     '\r': 1,
	     '\n': 1}
assert cont_dict.keys() == grade_dict.keys()

def review(card, mode):
	start = datetime.datetime.now()  # begin card timer

	# dynamically import the required reviewer module
	# a specifc reviewer is responsible for returning a key to the generic reviewer
	m = importlib.import_module('.'+card.reviewer, package = 'vinca.reviewers')
	key = m.review(card, mode)  # the reviewer gives back the key

	stop = datetime.datetime.now()
	elapsed_time = min(120, (stop - start).seconds)

	if key in ['a']:
		pass  #TODO card creation during review
	elif key in ['e','f','b','t','*']:
		card.edit(mode = key)
		return card.review(mode = 'preview')
	elif key in grade_dict.keys():
		cont = cont_dict[key]
		grade = grade_dict[key] if mode=='review' else 0
	else:
		raise ValueError(f'The key you pressed is not valid. Try {grade_dict.keys()}')
		grade, cont = 0, 0  # marks the card as seen and will instruct us to exit
	card.add_history(today, elapsed_time, grade)
	return cont

def default_str(card):
	m = importlib.import_module('.'+card.reviewer, package = 'vinca.reviewers')
	assert hasattr(m, 'default_str')
	return m.default_str(card)



from vinca.lib.ansi import ansi
import shutil

COLUMNS, LINES = shutil.get_terminal_size()
TAB_WIDTH = 8

# A Context Manager for the terminal's alternate screen
class AlternateScreen:

	def __init__(self):
		pass

	def __enter__(self):
		print(ansi['save_cursor'] + ansi['hide_cursor'], end='')
		print(ansi['save_screen'] + ansi['clear'] + ansi['move_to_top'], end='')

	def __exit__(self, *exception_args):
		print(ansi['restore_screen'],end='')
		print(ansi['show_cursor'] + ansi['restore_cursor'],end='')



from vinca import parsing, functions
from vinca.lib.classes import Card, Deck
import sys

# parse the command line arguments
parser = parsing.parser
args = parser.parse_args()
# sort decks_or_cards appropriately
for e in args.decks_or_cards:
	if type(e) is Card:
		args.cards.append(e)
	if type(e) is Deck:
		args.decks.append(e)	
# accept a file of newline separated cards
if not sys.stdin.isatty():
	for line in sys.stdin:
		id = line.strip()
		assert id.isdigit(), f'Bad card id {id}'
		args.cards.append(Card(id))
	
# run the specified function
func = args.func
func = getattr(functions, func)
func(args)



# this is the base editor for creation of simple front/back plaintext cards
# TODO no filename
#      no buffer titles
#      no tildes
import subprocess
from pathlib import Path
import shutil

vinca_path = Path(__file__).parent.parent.parent
tags_path = vinca_path / 'data' / 'tags.txt'

path = Path(__file__).parent
vimrc_path = path / 'vimrc'
# we create a few files to store file paths
# The user will specify media file paths in these files
# then the editor will copy the file specified by the filepath into the card
path_spec_image_front = path / 'image_front'
path_spec_image_back = path / 'image_back'
path_spec_audio_front = path / 'audio_front'
path_spec_audio_back = path / 'audio_back'
# create these temp files and delete previous contents
path_spec_image_front.open('w').close
path_spec_image_back.open('w').close
path_spec_audio_front.open('w').close
path_spec_audio_back.open('w').close

vimrc_modes = {'a': ['-c', 'startinsert'],
	       'e': [],
	       'f': [],
               'b': ['-c', '2 wincmd W'],
               't': ['-c', '3 wincmd W']}


def edit(card, mode):
	# possible modes are a, e, f, b
	# they stand for add, edit, front_edit, and back_edit
	implemented_modes = ['a','e','f','b','t']
	if mode not in implemented_modes:
		return

	if mode == 'a':
		card.make_auxfile('front')
		card.make_auxfile('back')

	assert 'front' in card.auxfiles
	assert 'back' in card.auxfiles

	# we are going to run vim...
	vim_cmd = ['vim']
	vim_cmd += [card.path/'front']

	# we are going to open each file window
	# one by one so that we can specify things well
	# this tells vim that we do not want to automatically
	# resize/equalize windows each time we open one
	# l will store a list of commands to open the window splits
	l = ['set noequalalways']
	# when we open a window split below we automatically move focus to that window
	l += ['set splitbelow']

	# open the back
	l += [f'split {card.path / "back"}']

	# 5 specifies that we want the new split to have 5 lines 
	# new splits are made by taking away space from the parent
	# therefore we start by giving the 'tags' window 5 lines
	# because the next split will take four lines (3 + status line)
	l += [f'5 split {card.path / "tags"}']
	l += [f'3 split {path_spec_image_back}']
	l += [f'1 split {path_spec_audio_back}']
	# vsplit stands for vertical split
	l += [f'vsplit {path_spec_audio_front}']
	# move up one window to draw the image_front split
	l += ['wincmd k']
	l += [f'vsplit {path_spec_image_front}']
	# bring focus back up to the top
	l += ['1 wincmd W']
	vim_cmd += ['-c',' | '.join(l)]  # vim uses | to join commands

	# using a vimrc file to make a few custom bindings...
	vim_cmd += ['-Nu', vimrc_path]
	# including tag autocompletion
	vim_cmd += ['-c', f'set dictionary={tags_path}']
	# with special options depending on the editing mode.
	vim_cmd += vimrc_modes[mode]
	# launch
	subprocess.run(vim_cmd)

	# finally process the media references
	for ref in ['image_front','image_back','audio_front','audio_back']:
		media_ref = (path / ref).read_text().strip()
		card_media = card.path / ref
		if media_ref == 'none' and card_media.exists():
			card_media.unlink()
		elif media_ref:
			source_file = Path( Path.cwd(), media_ref )
			try:
				assert source_file.exists(), f"The media file {source_file} could not be found"
				assert source_file.is_file(), f"{source_file} is not an image file."
			except:
				continue
			shutil.copy( media_ref, card_media )



import re

# ansi codes let us control the terminal just
# by printing special escape sequences
esc = '\033['
ansi = {'reset': esc+'0m',
        'bold': esc+'1m', 'b': esc+'1m',
        'bold_off': esc+'22m', 'b': esc+'1m',
        'light': esc+'2m', 'dim': esc+'2m',
        'italic': esc+'3m', 'it': esc+'3m',
        'underlined': esc+'4m', 'u': esc+'4m',
        'blink': esc+'5m', 'flash': esc+'5m', 'f': esc+'5m',
        'highlight': esc+'7m', 'hi': esc+'7m', 'reverse': esc+'7m',  'mark': esc+'7m',
        'hidden': esc+'8m', 'invisible': esc+'8m',
        'red': esc+'31m', 'r': esc+'31m',
        'green': esc+'32m', 'g': esc+'32m',
        'yellow': esc+'33m',
	'blue': esc+'34m',
	'reset_color': esc+'39m',
	# the next escape codes are more powerful
	'clear': esc+'2J',
	'move_to_top': esc+'H',  # home
	'hide_cursor': esc+'?25l',
	'show_cursor': esc+'?25h',
	'save_cursor': esc+'s',
	'restore_cursor': esc+'u',
	'save_screen': esc+'?47h',
	'restore_screen': esc+'?47l',
	'move_up_line': esc+'F',
	'clear_line': esc+'K',
	}

# we define a subclass of string
# ANSI codes count as characters normally.
# We want want to be able to subscript the string
# so that the length of the slice will be the number of LETTERS
# not the number of characters (which would include the ansi codes)
class AnsiStr(str):

	def __getitem__(self, subscript):
		# we want to split on ansi codes
		# and keep the ansi codes as members of the returned list
		codes = [c.replace('[','\[').replace('?','\?') for c in ansi.values()]
		split_pattern = '(' + '|'.join(codes) + ')'
		words_and_codes = re.split(split_pattern, self)

		# now we construct a list in which each element contains
		# exactly one letter and any number of ansi codes
		l=['']
		for wc in words_and_codes:
			if wc in ansi.values(): # append codes to preceding letter
				l[-1] += wc
			else: # add the letters as individual elements
				l += list(wc)
		# the first entry does not contain a letter so
		# we combine the first and second entries
		l = [l[0]+l[1]] + l[2:]

		if type(subscript) is int:
			return l[subscript] 
		elif type(subscript) is slice:
			return ''.join(l[subscript])



from pathlib import Path
from shutil import copytree, rmtree
from subprocess import run
import datetime
today = datetime.date.today()
study_day = today

import readchar  # 3rd party module for reading a character one at a time

from vinca.lib import classes
col = classes.collection  # TODO classes should not initialize vars
deckdict = classes.deckdict
from vinca.lib import browser
from vinca.lib.ansi import ansi
from vinca.lib.terminal import COLUMNS, LINES, TAB_WIDTH

vinca_path = Path(__file__) # /path/to/vinca/functions.py
vinca_path = vinca_path.parent # /path/to/vinca
cards_path = vinca_path / 'data' / 'cards'
decks_path = vinca_path / 'data' / 'decks.txt'
decks_vimrc = vinca_path / 'lib' / 'decks.vimrc'
tags_path = vinca_path / 'data' / 'tags.txt'


def study(args):
	queue = []
	if not args.decks and not args.cards:
		queue = browser.filter(col, due_only=True)
	for deck in args.decks:
		queue += browser.filter(col, deck=deck, due_only=True)
	queue += args.cards
	done_queue = []
	while queue:
		card = queue.pop()
		cont = card.review()
		if cont == -1:
			card.undo_history()
			queue.append(card)
			if not done_queue:
				break
			prev_card = done_queue.pop()
			prev_card.undo_history()
			queue.append(prev_card)
		if cont == 0:
			card.undo_history()
			break
		if cont == 1:
			card.schedule()
			done_queue.append(card)

# generators
def linear_add(args):
	new_card = classes.Card(create=True)
	new_card.make_config('linear','linear','base')
	new_card.add_tags([tag for deck in args.decks for tag in deck.tags_create])
	new_card.edit('a')
	new_card.schedule()
def one_liner(args):
	new_card = classes.Card(create=True)
	new_card.make_config('base','base','base')
	new_card.add_tags([tag for deck in args.decks for tag in deck.tags_create])
	# BAD: this reading should be implemented elsewhere
	(new_card.path/'front').write_text(input('Q:   '))
	(new_card.path/'back').write_text(input('A:   '))
	new_card.add_history(today, 0, 0)
	new_card.schedule()
def add_basic(args):
	new_card = classes.Card(create=True)
	new_card.make_config('base','base','base')
	new_card.add_tags([tag for deck in args.decks for tag in deck.tags_create])
	new_card.edit(mode='a')  # add mode
	cont = new_card.review(mode='preview')
	return new_card, cont
def add(args):
	new_card, cont = add_basic(args)
	new_card.schedule()
def add_many(args): #TODO
	prev_card = None
	while True:	
		new_card, cont = add_basic(args)
		new_card.schedule()
		if cont == -1:
			prev_card.review(mode='preview')  # TODO does not support multilevel
			prev_card.undo_history()  # we do not actually want to change the old card
		if cont == 0:
			break
		if cont == 1:
			prev_card = new_card
def image_cloze(args):
	pass
def statistics(args):
	for deck in args.decks:
		all_cards = browser.filter(col, decks=[deck])
		due_cards = browser.filter(all_cards, due_only=True)
		new_cards = browser.filter(all_cards, new_only=True)
		print(f'Total\t{len(all_cards)}')
		print(f'Due\t{len(due_cards)}')
		print(f'New\t{len(new_cards)}')
	for card in args.cards:
		print(f'\nCard #{card.id}')
		print(card)
		print(f'Tags: {" ".join(card.tags)}')
		print(f'Due: {card.due_date}')
		print('Date\t\tTime\tGrade')
		print(*[f'{date}\t{time}\t{grade}' for date, time, grade in card.history],sep='\n')
	if not args.cards and not args.decks:
		due_cards = browser.filter(col, due_only=True)
		new_cards = browser.filter(col, new_only=True)
		print(f'Total\t{len(col)}')
		print(f'Due\t{len(due_cards)}')
		print(f'New\t{len(new_cards)}')
def edit(args):
	if args.cards:
		for card in args.cards:
			card.edit()
		return
	vim_cmd = ['vim',decks_path] 
	vim_cmd += [f'+{args.decks[0].id + 1}'] if args.decks else []
	vim_cmd += ['-Nu',decks_vimrc]
	vim_cmd += ['-c',f'set dictionary={tags_path}']
	run(vim_cmd)
def delete(args):
	for deck in args.decks: 
		deckdict.delete(deck)
	for card in args.cards:
		if card.deletedQ:  #toggle
			card.undelete()
			continue
		card.delete()

# visual selection
cmd_dict = {'a':add,
	    'A':add_many,
	    'e':edit,
	    'x':delete,
	    's':study,
	    'S':statistics,
	    'l':linear_add,
	    '1':one_liner}
def visual_select(args, iterable):
	# allow for the selection of a card or deck
	n = len(iterable)
	sel = 0  # selected number
	print(ansi['hide_cursor'],end='')
	print('\n'*n,end='')  # move down n lines
	w = COLUMNS - TAB_WIDTH  # desired string width
	while True:
		print(ansi['move_up_line']*n,end='') # move up n lines
		for i, item in enumerate(iterable):
			print(ansi['reverse']*(i==sel) + str(item.id) + 
				'\t' + str(item)[:w] + ansi['reset'])
		k = readchar.readchar()  # get key
		sel = (sel + (k=='j') - (k=='k') ) % n
		if k == 'q' or k == readchar.key.ESC:
			print(ansi['show_cursor'],end='')
			return
		if k in cmd_dict.keys():
			if type(item) is classes.Deck: args.decks = [iterable[sel]]
			if type(item) is classes.Card: args.cards = [iterable[sel]]
			cmd_dict[k](args)
			if k == 'S':
				break
			print(ansi['hide_cursor'],end='')
def list_decks(args):
	assert deckdict, 'You have no decks! Use [e] to make a deck.'
	if args.mode=='visual':
		visual_select(args, deckdict.list)
		return
	print(*[f'{d.id}\t{d}' for d in deckdict.list], sep='\n')
def query(args):
	# we search the collection unless a list of cards has been piped
	matches = browser.query(args.cards if args.cards else col, args.pattern, fancy=True)
	if not matches:
		print('No matches.')
		return
	w = COLUMNS - TAB_WIDTH  # desired string width
	if args.mode=='visual':
		visual_select(args, matches)
		return
	for card in matches:
		print(str(card.id) + ('\t' + str(card)[:w]) * (not args.id_only))
def filter(args):
	unfiltered_cards = args.cards if args.cards else col
	filtered_cards = browser.filter(unfiltered_cards,
		**{param : getattr(args, param) for param in ['decks',
		'tags_include','tags_exclude','create_date_min','create_date_max',
		'seen_date_min','seen_date_max','due_date_min','due_date_max',
		'due_only','not_due_only','editor','reviewer','scheduler',
		'deleted_only','not_deleted_only','new_only','not_new_only']})
	w = COLUMNS - TAB_WIDTH
	if args.mode=='visual':
		visual_select(args, filtered_cards)
		return
	for card in filtered_cards:
		print(str(card.id) + ('\t' + str(card)[:w]) * (not args.id_only))

def purge(args):
	for card in browser.filter(col, deleted_only=True):
		rmtree(card.path)
# backup / export
def backup(args):
	backup_cards = [] if args.cards or args.decks else col
	backup_cards += [c for d in args.decks for c in browser.filter(col, deck=d)]
	backup_cards += args.cards

	for card in backup_cards:
		copytree(card.path, args.backup_dest / str(card.id))
def import_collection(args):
	# TODO implement a hash comparison to prevent card duplication
	if args.overwrite:
		rmtree(cards_path)
		copytree(args.import_path, cards_path)
		return
	for new_id,card_path in enumerate(args.import_path.iterdir(), start=max(col.id_list) + 1):
		copytree(card_path, cards_path / str(new_id))



# A basic browser
import datetime
TODAY = datetime.date.today()
DAY = datetime.timedelta(days=1)
import re
from vinca.lib.ansi import ansi, AnsiStr

def filter(cards,
	   tags_include=None, tags_exclude=None, decks=None,
	   create_date_min=None, create_date_max=None,
	   seen_date_min=None, seen_date_max=None,
	   due_date_min=None, due_date_max=None,
	   due_only=False, not_due_only = False,
	   editor=None, reviewer=None, scheduler=None,
	   deleted_only=False, not_deleted_only=False,
	   new_only=False, not_new_only=False):
	
	# initializing a parameter to [] will cause bugs
	# since lists are mutable and parameters are created
	# when the function is defined (not when called)
	decks = decks if decks else []
	tags_include = tags_include if tags_include else []
	tags_exclude = tags_exclude if tags_exclude else []
	tags_include += [t for d in decks for t in d.tags_include]
	tags_exclude += [t for d in decks for t in d.tags_exclude]

	matches = []
	for card in cards:
		# tags
		if tags_include and not any([t in tags_include for t in card.tags]):
			continue
		if tags_exclude and any([t in tags_exclude for t in card.tags]):
			continue
		# dates
		if create_date_min and card.create_date < create_date_min:
			continue
		if create_date_max and card.create_date > create_date_max:
			continue
		if seen_date_min and card.seen_date < seen_date_max:
			continue
		if seen_date_max and card.seen_date > seen_date_max:
			continue
		if due_date_min and card.due_date < due_date_min:
			continue
		if due_date_max and card.due_date > due_date_max:
			continue
		# config
		if editor and card.editor != editor:
			continue
		if reviewer and card.reviewer != reviewer:
			continue
		if scheduler and card.scheduler != scheduler:
			continue
		# due / new / deleted
		if due_only and card.due_date > TODAY:
			continue
		if not_due_only and card.due_date <= TODAY:
			continue
		if new_only and not card.newQ:
			continue
		if not_new_only and card.newQ:
			continue
		if deleted_only and not card.deletedQ:
			continue
		if not_deleted_only and card.deletedQ:
			continue

		matches.append(card)

	return matches

def query(cards, pattern, fancy=False):
	# This lets us search our collection for a regular expression
	# (Which would be better than anki!)

	p = re.compile(f'({pattern})')  # wrap in parens to create regex group \1
	matches = [c for c in cards if p.search(str(c))]
	if fancy:
		for c in matches:
			c.str = p.sub(f'{ansi["red"]}{ansi["bold"]}\\1' +
				      f'{ansi["reset_color"]}{ansi["bold_off"]}',str(c))
			c.str = AnsiStr(c)  # AnsiStr class lets slicing ignore ansi chars
	return matches
